-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local table = _G.table

local tsort = table.sort

-------------------------------------------------------------------------------
-- Localized Blizzard API
-------------------------------------------------------------------------------
local build = select(2, GetBuildInfo())
build = tonumber(build)

local GetQuestDifficultyColor = (build >= 10048) and _G.GetQuestDifficultyColor or _G.GetDifficultyColor

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local MODNAME = "GearGauge"
local GearGauge = LibStub("AceAddon-3.0"):GetAddon(MODNAME)

local QTip = LibStub("LibQTip-1.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

-------------------------------------------------------------------------------
-- Constants.
-------------------------------------------------------------------------------
local ORDERED_CLASSES = GearGauge.ORDERED_CLASSES
local CLASS_COLORS = GearGauge.CLASS_COLORS
local MY_COLOR = GearGauge.player_color

-------------------------------------------------------------------------------
-- Variables.
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- Local functions.
-------------------------------------------------------------------------------
local function TimeStr(seconds)
	local tm = seconds
	local timeval = tostring(math.floor(tm % 60)).."s"
	local val

	if (tm >= 60) then
		tm = tm / 60
		val = math.floor(tm % 60)

		if val > 0 then	timeval = tostring(val).."m "..timeval end

		if (tm >= 60) then
			tm = tm / 60
			val = math.floor(tm % 24)

			if val > 0 then	timeval = tostring(val).."h "..timeval end

			if (tm >= 24) then
				tm = tm / 24
				timeval = tostring(math.floor(tm)).."d "..timeval
			end
		end
	end
	return timeval
end

-------------------------------------------------------------------------------
-- List Panel
-------------------------------------------------------------------------------
local def_col, def_bg_col = TOOLTIP_DEFAULT_COLOR, TOOLTIP_DEFAULT_BACKGROUND_COLOR
local list_font = CreateFont(MODNAME.."ListFont")
list_font:SetFont("Fonts\\FRIZQT__.TTF", 10)

-------------------------------------------------------------------------------
-- List Panel variables.
-------------------------------------------------------------------------------
local list_entries = {}
local name_list = {}

-- Sort in ascending order: If one and two are equal, return descending order in name_list.
local function SortAsc(one, two, field)
	if (not one[field] and not two[field]) or one[field] == two[field] then
		return name_list[one] < name_list[two]
	elseif not two[field] then
		return true
	elseif not one[field] then
		return false
	else
		return one[field] > two[field]
	end
end

-- Sort in descending order: If one and two are equal, return descending order in name_list.
local function SortDesc(one, two, field)
	if (not one[field] and not two[field]) or one[field] == two[field] then
		return name_list[one] < name_list[two]
	elseif not one[field] then
		return true
	elseif not two[field] then
		return false
	else
		return one[field] < two[field]
	end
end

local SORT_FUNCS = {
	["ClassSortDown"] = function(one, two)
				    local one_cl, two_cl = ORDERED_CLASSES[one.class], ORDERED_CLASSES[two.class]

				    if one_cl == two_cl then
					    return name_list[one] < name_list[two]
				    end
				    return one_cl < two_cl
			    end,
	["ClassSortUp"] = function(one, two)
				  local one_cl, two_cl = ORDERED_CLASSES[one.class], ORDERED_CLASSES[two.class]

				  if one_cl == two_cl then
					  return name_list[one] < name_list[two]
				  end
				  return one_cl > two_cl
			  end,
	["FactionSortDown"] = function(one, two) return SortDesc(one, two, "faction") end,
	["FactionSortUp"] = function(one, two) return SortAsc(one, two, "faction") end,
	["GuildSortDown"] = function(one, two) return SortDesc(one, two, "guild") end,
	["GuildSortUp"] = function(one, two) return SortAsc(one, two, "guild") end,
	["ItemsSortDown"] = function(one, two) return SortDesc(one, two, "equip_level") end,
	["ItemsSortUp"] = function(one, two) return SortAsc(one, two, "equip_level") end,
	["LevelSortDown"] = function(one, two) return SortDesc(one, two, "level") end,
	["LevelSortUp"] = function(one, two) return SortAsc(one, two, "level") end,
	["NameSortDown"] = function(one, two) return name_list[one] < name_list[two] end,
	["NameSortUp"] = function(one, two) return name_list[one] > name_list[two] end,
	["RaceSortDown"] = function(one, two) return SortDesc(one, two, "race") end,
	["RaceSortUp"] = function(one, two) return SortAsc(one, two, "race") end,
	["RatingSortDown"] = function(one, two) return SortDesc(one, two, "rating") end,
	["RatingSortUp"] = function(one, two) return SortAsc(one, two, "rating") end,
	["ScanSortDown"] = function(one, two) return SortAsc(one, two, "scanned") end,
	["ScanSortUp"] = function(one, two) return SortDesc(one, two, "scanned") end,
}
local game_font_color = string.format("|cff%2x%2x%2x", NORMAL_FONT_COLOR.r * 255, NORMAL_FONT_COLOR.g * 255, NORMAL_FONT_COLOR.b * 255)
local prev_name, prev_sort

-------------------------------------------------------------------------------
-- List Panel helper functions
-------------------------------------------------------------------------------
local panel, list, menu_pane

local function GenerateList(match_str)
	local realm = GearGauge.db.realm

	wipe(list_entries)
	wipe(name_list)

	if match_str then
		local target, l_name, l_class, l_guild, l_race

		match_str = match_str:lower()

		for entry in pairs(realm.entries) do
			target = realm.entries[entry]
			l_name = string.lower(entry)
			l_class = string.lower(GearGauge.db.global.localized_classes[ORDERED_CLASSES[target.class]])
			l_guild = target.guild and string.lower(target.guild) or nil
			l_race = string.lower(target.race)

			if (string.match(l_name, match_str) or string.match(l_class, match_str) or
			    (l_guild and string.match(l_guild, match_str)) or string.match(l_race, match_str)) then
				name_list[target] = entry
				tinsert(list_entries, target)
			end
		end
	else
		for entry in pairs(realm.entries) do
			name_list[realm.entries[entry]] = entry
			tinsert(list_entries, realm.entries[entry])
		end
	end
end

local function ScanMembers(type)
	if InCombatLockdown() then
		GearGauge:Print(ERR_NOT_IN_COMBAT)
		return
	end
	local members = (type == "party") and GetNumPartyMembers() or GetNumRaidMembers()

	if members == 0 then
		GearGauge:Print((type == "party") and ERR_NOT_IN_GROUP or ERR_NOT_IN_RAID)
		return
	end
	local avg_lev, avg_rating = 0, 0
	local unit, target, tmp_lev, tmp_rating

	wipe(list_entries)
	wipe(name_list)

	for i = 1, members, 1 do
		unit = type..i
		tmp_rating, tmp_lev = GearGauge:Scan(unit)
		target = GearGauge.db.realm.entries[UnitName(unit)]

		-- If the entry doesn't exist, it wasn't in the list previously and the target was too far away to scan.
		if target then
			name_list[target] = UnitName(unit)
			tinsert(list_entries, target)

			if tmp_rating and tmp_level then
				avg_rating = avg_rating + tmp_rating
				avg_lev = avg_lev + tmp_lev
			end
		end
	end
	GearGauge.panel:Refresh()
	--		GearGauge:Print(string.format("Average rating: %d, Average equipment level: %d", avg_rating, avg_lev))
end

local function MenuPane_Click(cell, arg, button)
	if not arg then
		GearGauge.panel:Hide()
		return
	end

	if arg == "all" then
		GearGauge.panel:Refresh(true)
	elseif arg == "party" or arg == "raid" then
		ScanMembers(arg)
	elseif arg == "guild" then
		panel:Hide()
		GenerateList(GetGuildInfo("player"))
		GearGauge.panel:Open()
	else
		local dir = (button == "LeftButton") and "Down" or "Up"

		prev_sort = arg..dir
		panel:Hide()
		GearGauge.panel:Open(prev_sort, false)
	end
end

local highlight = CreateFrame("Frame", nil, UIParent)
highlight:SetFrameStrata("TOOLTIP")
highlight:Hide()

highlight._texture = highlight:CreateTexture(nil, "OVERLAY")
highlight._texture:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
highlight._texture:SetBlendMode("ADD")
highlight._texture:SetAllPoints(highlight)

local function ListOnEnter(frame)
	highlight:SetParent(frame)
	highlight:SetAllPoints(frame)
	highlight:Show()
end

local function ListOnLeave(frame)
	highlight:Hide()
	highlight:ClearAllPoints()
	highlight:SetParent(nil)
end

local function ListOnMouseUp(frame, button)
	if prev_name == frame.arg then
		prev_name = nil
		GearGauge.tooltip = QTip:Release(GearGauge.tooltip)
		return
	end

	if IsAltKeyDown() and button == "RightButton" then
		panel:Hide()
		RemoveEntry(frame.arg, true)
		GearGauge.panel:Open(prev_sort, true)
		return
	end
	prev_name = frame.arg
	GearGauge:ShowPlayerGear(frame.arg, panel)		-- FIXME: This shouldn't have to be called twice to get correct gem/enchant information.
	GearGauge:ShowPlayerGear(frame.arg, panel)
end

local function SetBorders(frame)
	local left = frame:CreateTexture(nil, "BACKGROUND")
	left:SetTexture(0.4, 0.4, 0.4)
	left:SetWidth(1)
	left:SetPoint("TOPLEFT")
	left:SetPoint("BOTTOMLEFT")

	local right = frame:CreateTexture(nil, "BACKGROUND")
	right:SetTexture(0.4, 0.4, 0.4)
	right:SetWidth(1)
	right:SetPoint("TOPRIGHT")
	right:SetPoint("BOTTOMRIGHT")

	local top = frame:CreateTexture(nil, "BACKGROUND")
	top:SetTexture(0.4, 0.4, 0.4)
	top:SetHeight(1)
	top:SetPoint("TOPLEFT")
	top:SetPoint("TOPRIGHT")

	local bottom = frame:CreateTexture(nil, "BACKGROUND")
	bottom:SetTexture(0.4, 0.4, 0.4)
	bottom:SetHeight(1)
	bottom:SetPoint("BOTTOMLEFT")
	bottom:SetPoint("BOTTOMRIGHT")

	local bg = frame:CreateTexture(nil, "BACKGROUND")
	bg:SetTexture(0, 0, 0, 1)
	bg:SetPoint("TOPLEFT", 1, -1)
	bg:SetPoint("BOTTOMRIGHT", -1, 1)
end

-------------------------------------------------------------------------------
-- Main Panel
-------------------------------------------------------------------------------
local PANEL_NAME = MODNAME.."Panel"
panel = CreateFrame("Frame", PANEL_NAME, UIParent)
panel:SetFrameStrata("DIALOG")
panel:SetBackdrop({
			  bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
			  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			  tile = true, tileSize = 32, edgeSize = 32,
			  insets = { left = 11, right = 12, top = 12, bottom = 11 }
		  })
panel:SetBackdropColor(def_col.r, def_col.g, def_col.b, 1)
panel:SetBackdropBorderColor(def_col.r, def_col.g, def_col.b)
panel:SetAlpha(1)
panel:SetWidth(750)
panel:SetHeight(322)
panel:SetPoint("CENTER", UIParent, 0, 0)
panel:Hide()
panel:SetMovable(true)
panel:SetClampedToScreen(true)
panel:SetScript("OnHide", function() GearGauge.tooltip = QTip:Release(GearGauge.tooltip) end)
panel:EnableMouseWheel(true)
panel:SetScript("OnMouseWheel", function(self, delta)
					local slider = self.slider
					local currentValue = slider:GetValue()
					local minValue, maxValue = slider:GetMinMaxValues()

					if delta < 0 and currentValue < maxValue then
						slider:SetValue(min(maxValue, currentValue + 10))
					elseif delta > 0 and currentValue > minValue then
						slider:SetValue(max(minValue, currentValue - 10))
					end
				end)

tinsert(UISpecialFrames, PANEL_NAME)
GearGauge.panel = panel

local title_box = CreateFrame("Frame", nil, panel)
title_box:SetFrameStrata("FULLSCREEN_DIALOG")
title_box:EnableMouse(true)
title_box:SetMovable(true)
title_box:RegisterForDrag("LeftButton")
title_box:SetScript("OnDragStart", function() panel:StartMoving() end)
title_box:SetScript("OnDragStop", function() panel:StopMovingOrSizing() end)

local title_bg = title_box:CreateTexture(nil, "ARTWORK")
title_bg:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
title_bg:SetPoint("CENTER", panel, "TOP", 0, -17)
title_bg:SetWidth(230)
title_bg:SetHeight(56)

local title_text = title_box:CreateFontString(nil, "ARTWORK", "GameFontNormal")
title_text:SetPoint("TOP", title_bg, "TOP", 0, -11)
title_text:SetText(MODNAME)

title_box:SetAllPoints(title_bg)

-------------------------------------------------------------------------------
-- List pane.
-------------------------------------------------------------------------------
list = CreateFrame("Frame", MODNAME.."PanelList", panel)
list:SetBackdrop(GameTooltip:GetBackdrop())
list:SetBackdropColor(GameTooltip:GetBackdropColor())
list:SetBackdropBorderColor(GameTooltip:GetBackdropBorderColor())
list:SetScale(GameTooltip:GetScale())
list:SetAlpha(1)
list:SetFrameStrata("DIALOG")
list:SetHeight(305)
list:SetWidth(605)

-------------------------------------------------------------------------------
-- List pane scrollbar.
-------------------------------------------------------------------------------
local EDGE, LINE_HEIGHT, LINE_GAP, GAP = 10, 11, 2, 6
local NUM_LINES = 21

local lines = {}
local cell_widths = {}

local cell_names = {
	[1] = "faction",	[2] = "name",		[3] = "level",
	[4] = "rating",		[5] = "equip_level",	[6] = "race",
	[7] = "class",		[8] = "guild",		[9] = "scanned"
}

-- Normalize cell widths for the panel list. Creates a throwaway frame and fontstring - should fix this at some point.
function GearGauge:InitializeList()
	for idx, name in ipairs(cell_names) do
		cell_widths[name] = 0
	end
	local now = time()
	local class_english, class_name

	local f = CreateFrame("Frame", nil, UIParent)
	local fs = f:CreateFontString()
	fs:SetFontObject(list_font)

	for plr in pairs(self.db.realm.entries) do
		local entry = self.db.realm.entries[plr]

		for k, v in ipairs(cell_names) do
			if v ~= "faction" then		-- We skip faction because we're not using the faction string in the list.
				if v == "scanned" then
					fs:SetText(TimeStr(now - entry.scanned))
				elseif v == "name" then
					fs:SetText(plr)
				elseif v == "class" then
					class_english = ORDERED_CLASSES[entry.class]
					class_name = GearGauge.db.global.localized_classes[class_english] or class_english

					if not class_name or not CLASS_COLORS[class_english] then
						fs:SetText(entry.class)
					else
						fs:SetText("|cff"..CLASS_COLORS[class_english]..class_name.."|r")
					end
				else
					fs:SetText(tostring(entry[v]))
				end

				local width = fs:GetStringWidth()

				if width > cell_widths[v] then
					cell_widths[v] = width
				end
			end
		end
	end
	self.InitializeList = nil
end

do
	local function SetCellText(self, value)
		local parent = self:GetParent()
		local fs = self._string

		fs:SetJustifyH("LEFT")
		fs:SetText(tostring(value))
		fs:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
		fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
		fs:Show()

		local width = fs:GetStringWidth()
		local name = self._name

		if width > cell_widths[name] then
			cell_widths[name] = width
		else
			width = cell_widths[name]
		end
		self:SetWidth(width)

		local new_width = parent:GetWidth() + width
		parent:SetWidth(new_width)

		local height = fs:GetHeight()
		self:SetHeight(height)

		if height > parent.height then
			parent.height = height
			parent:SetHeight(height)
		end
	end
	local anchor

	for i = 1, NUM_LINES do
		local line = CreateFrame("Frame", nil, list)
		line.height = 1

		if anchor then
			line:SetPoint("TOP", anchor, "BOTTOM", 0, -LINE_GAP)
		else
			line:SetPoint("TOP", list, 0, -10)
		end
		line:SetPoint("LEFT", list, EDGE/2, 0)
		line:SetPoint("RIGHT", list, -EDGE/2 - 22, 0)

		local cell, relative

		for idx, name in ipairs(cell_names) do
			cell = CreateFrame("Frame", nil, line)
			cell:SetPoint("TOP", line)
			cell:SetPoint("BOTTOM", line)

			if relative then
				cell:SetPoint("LEFT", relative, "RIGHT", GAP, 0)
			else
				cell:SetPoint("LEFT", line)
			end

			local fstring = cell:CreateFontString()
			fstring:SetFontObject(list_font)

			cell._string = fstring
			cell._name = name
			cell.SetText = SetCellText
			cell:Show()

			line[name] = cell
			relative = cell
		end
		anchor = line
		lines[i] = line
	end
end	-- do

local scrollbar = CreateFrame("Slider", nil, list)
scrollbar:SetWidth(12)
scrollbar:SetOrientation("VERTICAL")
scrollbar:SetBackdrop({
			      bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
			      edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
			      tile = true, tileSize = 8, edgeSize = 8,
			      insets = { left = 3, right = 3, top = 3, bottom = 3 }
		      })
scrollbar:SetThumbTexture("Interface\\Buttons\\UI-SliderBar-Button-Vertical")
scrollbar:SetMinMaxValues(0, 1)
scrollbar:SetValueStep(1)

local function ShowList(position)
	local now = time()

	if #list_entries <= NUM_LINES then
		scrollbar:Hide()
	else
		scrollbar:Show()
	end

	for i, line in pairs(lines) do
		local entry = list_entries[i + position]

		if entry then
			local enemy = entry.faction ~= player_faction
			local diff = GetQuestDifficultyColor(entry.level)

			line.faction:SetText("|TInterface\\GROUPFRAME\\UI-Group-PVP-"..entry.faction..":16:16|t")
			line.name:SetText(name_list[entry])
			line.arg = name_list[entry]

			line:EnableMouse(true)
			line:SetScript("OnMouseUp", ListOnMouseUp)
			line:SetScript("OnEnter", ListOnEnter)
			line:SetScript("OnLeave", ListOnLeave)

			line.level:SetText(string.format("|cff%2x%2x%2x%d|r", diff.r * 255, diff.g * 255, diff.b * 255, entry.level))
			line.rating:SetText(entry.rating or 0)
			line.equip_level:SetText(entry.equip_level or 0)
			line.race:SetText(entry.race)

			local class_english = ORDERED_CLASSES[entry.class]
			local class_name = GearGauge.db.global.localized_classes[class_english] or class_english

			if not class_name or not CLASS_COLORS[class_english] then
				line.class:SetText(entry.class)
			else
				line.class:SetText("|cff"..CLASS_COLORS[class_english]..class_name.."|r")
			end
			line.guild:SetText(entry.guild or "")
			line.scanned:SetText(TimeStr(now - entry.scanned))
			line:Show()
		else
			line.arg = nil
			line:EnableMouse(false)
			line:SetScript("OnMouseUp", nil)
			line:SetScript("OnEnter", nil)
			line:SetScript("OnLeave", nil)
			line:Hide()
		end
	end
end

local function GenerateMenuPane()
	menu_pane:Clear()
	menu_pane:AddHeader(L["SORT_ENTRIES_STR"])
	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)

	local line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..CLASS.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "ClassSort")
 
	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..FACTION.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "FactionSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..GUILD.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "GuildSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..ITEMS.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "ItemsSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..LEVEL.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "LevelSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..NAME.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "NameSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..RACE.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "RaceSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..RATING.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "RatingSort")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..L["TIMESTAMP"].."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "ScanSort")

	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)

	menu_pane:AddHeader(DISPLAY)
	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..PARTY.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "party")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..RAID.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "raid")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..GUILD.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "guild")

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, game_font_color..ALL.."|r")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, "all")

	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)
	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, 1, game_font_color..CLOSE.."|r", "CENTER")
	menu_pane:SetCellScript(line, column, "OnMouseUp", MenuPane_Click, nil)

	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)
	menu_pane:AddSeparator(1, MY_COLOR.r, MY_COLOR.g, MY_COLOR.b)

	line, column = menu_pane:AddLine()
	menu_pane:SetCell(line, column, string.format(L["PLAYER_COUNT_STR"], #list_entries), "CENTER")
end

scrollbar:SetScript("OnValueChanged",
		    function(self, value, ...)
			    local offset = math.floor(value)

			    ShowList(offset)
			    GenerateMenuPane()
		    end)
panel.slider = scrollbar
-------------------------------------------------------------------------------
-- Menu pane.
-------------------------------------------------------------------------------
menu_pane = QTip:Acquire(MODNAME.."MenuPanel", 1, "CENTER")
menu_pane:SetParent(panel)
menu_pane:SetFrameStrata("DIALOG")
menu_pane:SetFont(list_font)
menu_pane:SetWidth(130)

-------------------------------------------------------------------------------
-- Search input box.
-------------------------------------------------------------------------------
local search_box = CreateFrame("EditBox", nil, panel)
search_box:SetWidth(126)
search_box:SetHeight(20)
search_box:SetMaxLetters(80)
search_box:SetAutoFocus(false)
search_box:SetFontObject("ChatFontSmall")
search_box:SetScript("OnEditFocusGained", function(self) self:HighlightText() end)
search_box:SetScript("OnEditFocusLost", function(self) self:HighlightText(0, 0) end)
search_box:SetScript("OnEnterPressed", function(self)
					       GearGauge.panel:Hide()
					       GenerateList(self:GetText())
					       GearGauge.panel:Open()
				       end)
search_box:SetScript("OnEscapePressed", function(self) self:ClearFocus() end)
SetBorders(search_box)

-------------------------------------------------------------------------------
-- Panel anchoring.
-------------------------------------------------------------------------------
list:SetPoint("TOPLEFT", panel, "TOPLEFT", 10, -10)
list:SetPoint("TOPRIGHT", menu_pane, "TOPLEFT", 3, 0)

scrollbar:SetPoint("TOPRIGHT", list, "TOPRIGHT", -10, -10)
scrollbar:SetPoint("BOTTOMRIGHT", list, "BOTTOMRIGHT", -10, 10)

menu_pane:SetPoint("TOPRIGHT", panel, "TOPRIGHT", -9, -10)
menu_pane:SetPoint("BOTTOMRIGHT", search_box, "TOPRIGHT", 0, 0)

search_box:SetPoint("BOTTOMRIGHT", panel, "BOTTOMRIGHT", -11, 11)

if TipTac and TipTac.AddModifiedTip then
	TipTac:AddModifiedTip(list, true)
	TipTac:AddModifiedTip(menu_pane, true)
end

-------------------------------------------------------------------------------
-- Main list panel function.
-------------------------------------------------------------------------------
function panel:Open(sort_func, generate)
	if self:IsVisible() then
		self:Hide()
		return
	end

	if generate then
		GenerateList()
	end

	local show_enemies = GearGauge.db.global.scanning.enemies
	if not sort_func or sort_func == "" or not SORT_FUNCS[sort_func] then
		sort_func = "NameSortDown"
	end
	tsort(list_entries, SORT_FUNCS[sort_func])

	ShowList(0)
	scrollbar:SetMinMaxValues(0, math.max(0, #list_entries - #lines))
	scrollbar:SetValue(0)

	GenerateMenuPane()
	self:Show()
	list:Show()
	menu_pane:Show()
end

-- Refresh the list if the panel is visible.
function panel:Refresh(generate)
	if self:IsVisible() then
		self:Hide()
		self:Open(prev_sort, generate)
	end
end
